// Module: drag.js
// Uses the built-in HTML support for dragging and dropping elements.
// When an element is draggable, its ID is saved as data for the drag
// When dropped, the ID is passed to the drop handler.
//
// Since we are using jQuery, we pass data to the handler using the data
// property of jQuery Events.
//
// For the drop handler we use a variable at top module level to pass the dragged element.

import $ from "./jquery.module.js";

let $draggedElement = null;

function makeDraggable({element, stop, moveElement}) {
    // Optional arguments:
    //      stop: handler,     -- If defined, a handler to be called for the event "dragend"
    //      moveElement: true, -- If defined and true, allows the mouse to move the element while dragging.
    //                            In this case, the body is set as droppable to prevent the drag operation
    //                            from snapping back when finished.
    const $element = element instanceof $ ? element : $(element);
    $element.attr("draggable", "true").on("dragstart", {$element}, function(evt) {
        console.log("Called dragstart");
        // evt.originalEvent.dataTransfer.setData("text/plain", this.id);
        $draggedElement = evt.data.$element;
    });

    if (stop) {
        $element.on("dragend", {$element}, stop);
    }

    if (moveElement) {
        $element.on("dragstart",  startMoveElement);
        makeDroppable({element: "body", dropHandler: () => {}}); // Make body droppable but don't do anything
    }
}

function makeDroppable({element, dropHandler}) {
    const $element = element instanceof $ ? element : $(element);
    $element
        .on("dragover", (evt) => evt.preventDefault())
        .on("drop", function(evt) {
            evt.preventDefault();
            // const data = evt.originalEvent.dataTransfer.getData("text/plain");
            // console.log("Data = " + data);
            dropHandler({$draggedElement, $dropElement: $element});
            $draggedElement = null;
        });
}

function startMoveElement(evt) {
    console.log("Called startMoveElement");
    let $el = $(evt.currentTarget);
    let offset = {x: parseFloat($el.css("left")) - evt.pageX, y: parseFloat($el.css("top")) - evt.pageY};
    $el.on("drag",  offset, moveElement);
}

function moveElement(evt) {
    // For some reason we get called one more time when the mouse is released with pageX and pageY set to zero, so we
    // have to check for that
    let pageX = evt.pageX;
    let pageY = evt.pageY;
    if (pageX === 0 && pageY === 0) {
        return;
    }
    let $el = $(evt.currentTarget);
    let offset = evt.data;
    $el.css("left", offset.x + pageX);
    $el.css("top",  offset.y + pageY);
}

export {makeDraggable, makeDroppable};